package sun.invoke.util;

import java.util.EnumMap;
import java.lang.invoke.MethodType;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;

public class ValueConversions
{
    private static final Class<?> THIS_CLASS;
    private static final MethodHandles.Lookup IMPL_LOOKUP;
    private static final ValueConversions.WrapperCache[] UNBOX_CONVERSIONS;
    private static final Integer ZERO_INT;
    private static final Integer ONE_INT;
    private static final ValueConversions.WrapperCache[] BOX_CONVERSIONS;
    private static final ValueConversions.WrapperCache[] CONSTANT_FUNCTIONS;
    private static final MethodHandle CAST_REFERENCE;
    private static final MethodHandle IGNORE;
    private static final MethodHandle EMPTY;
    private static final ValueConversions.WrapperCache[] CONVERT_PRIMITIVE_FUNCTIONS;
    
    public ValueConversions() {
    }
    
    private static ValueConversions.WrapperCache[] newWrapperCaches(int n) {
        ValueConversions.WrapperCache[] array = new ValueConversions.WrapperCache[n];
        for (int i = 0; i < n; ++i) {
            array[i] = new ValueConversions.WrapperCache();
        }
        return array;
    }
    
    static int unboxInteger(Integer n) {
        return (int)n;
    }
    
    static int unboxInteger(Object o, boolean b) {
        if ((o instanceof Integer)) {
            return (int)o;
        }
        return primitiveConversion(Wrapper.INT, o, b).intValue();
    }
    
    static byte unboxByte(Byte b) {
        return (byte)b;
    }
    
    static byte unboxByte(Object o, boolean b) {
        if ((o instanceof Byte)) {
            return (byte)o;
        }
        return primitiveConversion(Wrapper.BYTE, o, b).byteValue();
    }
    
    static short unboxShort(Short n) {
        return (short)n;
    }
    
    static short unboxShort(Object o, boolean b) {
        if ((o instanceof Short)) {
            return (short)o;
        }
        return primitiveConversion(Wrapper.SHORT, o, b).shortValue();
    }
    
    static boolean unboxBoolean(Boolean b) {
        return (boolean)b;
    }
    
    static boolean unboxBoolean(Object o, boolean b) {
        if ((o instanceof Boolean)) {
            return (boolean)o;
        }
        return (primitiveConversion(Wrapper.BOOLEAN, o, b).intValue() & 0x1) != 0x0;
    }
    
    static char unboxCharacter(Character c) {
        return (char)c;
    }
    
    static char unboxCharacter(Object o, boolean b) {
        if ((o instanceof Character)) {
            return (char)o;
        }
        return (char)primitiveConversion(Wrapper.CHAR, o, b).intValue();
    }
    
    static long unboxLong(Long n) {
        return (long)n;
    }
    
    static long unboxLong(Object o, boolean b) {
        if ((o instanceof Long)) {
            return (long)o;
        }
        return primitiveConversion(Wrapper.LONG, o, b).longValue();
    }
    
    static float unboxFloat(Float n) {
        return (float)n;
    }
    
    static float unboxFloat(Object o, boolean b) {
        if ((o instanceof Float)) {
            return (float)o;
        }
        return primitiveConversion(Wrapper.FLOAT, o, b).floatValue();
    }
    
    static double unboxDouble(Double n) {
        return (double)n;
    }
    
    static double unboxDouble(Object o, boolean b) {
        if ((o instanceof Double)) {
            return (double)o;
        }
        return primitiveConversion(Wrapper.DOUBLE, o, b).doubleValue();
    }
    
    private static MethodType unboxType(Wrapper wrapper, int n) {
        if (n == 0) {
            return MethodType.methodType(wrapper.primitiveType(), wrapper.wrapperType());
        }
        return MethodType.methodType(wrapper.primitiveType(), Object.class, new Class[] { Boolean.TYPE });
    }
    
    private static MethodHandle unbox(Wrapper wrapper, int n) {
        ValueConversions.WrapperCache wrapperCache = ValueConversions.UNBOX_CONVERSIONS[n];
        MethodHandle value = wrapperCache.get(wrapper);
        if (value != null) {
            return value;
        }
        switch (wrapper) {
            case OBJECT:
            case VOID: {
                throw new IllegalArgumentException("unbox " + wrapper);
            }
            default: {
                String string = "unbox" + wrapper.wrapperSimpleName();
                MethodType unboxType = unboxType(wrapper, n);
                MethodHandle target;
                try {
                    target = ValueConversions.IMPL_LOOKUP.findStatic(ValueConversions.THIS_CLASS, string, unboxType);
                }
                catch (ReflectiveOperationException ex) {
                    target = null;
                }
                if (target != null) {
                    if (n > 0) {
                        target = MethodHandles.insertArguments(target, 1, new Object[] { Boolean.valueOf(n != 2) });
                    }
                    if (n == 1) {
                        target = target.asType(unboxType(wrapper, 0));
                    }
                    return wrapperCache.put(wrapper, target);
                }
                else {
                    throw new IllegalArgumentException("cannot find unbox adapter for " + wrapper + ((n <= 1) ? " (exact)" : ((n == 3) ? " (cast)" : "")));
                }
            }
        }
    }
    
    public static MethodHandle unboxExact(Wrapper wrapper) {
        return unbox(wrapper, 0);
    }
    
    public static MethodHandle unboxExact(Wrapper wrapper, boolean b) {
        return unbox(wrapper, b ? 0 : 1);
    }
    
    public static MethodHandle unboxWiden(Wrapper wrapper) {
        return unbox(wrapper, 2);
    }
    
    public static MethodHandle unboxCast(Wrapper wrapper) {
        return unbox(wrapper, 3);
    }
    
    public static Number primitiveConversion(Wrapper wrapper, Object obj, boolean b) {
        if (obj == null) {
            if (!b) {
                return null;
            }
            return ValueConversions.ZERO_INT;
        }
        else {
            Number value;
            if ((obj instanceof Number)) {
                value = (Number)obj;
            }
            else if ((obj instanceof Boolean)) {
                value = (((boolean)obj) ? ValueConversions.ONE_INT : ValueConversions.ZERO_INT);
            }
            else if ((obj instanceof Character)) {
                value = Integer.valueOf((int)(char)obj);
            }
            else {
                value = (Number)obj;
            }
            Wrapper wrapperType = Wrapper.findWrapperType(obj.getClass());
            if (wrapperType == null || (!b && !wrapper.isConvertibleFrom(wrapperType))) {
                return (Number)wrapper.wrapperType().cast(obj);
            }
            return value;
        }
    }
    
    public static int widenSubword(Object o) {
        if ((o instanceof Integer)) {
            return (int)o;
        }
        if ((o instanceof Boolean)) {
            return fromBoolean((boolean)o);
        }
        if ((o instanceof Character)) {
            return (char)o;
        }
        if ((o instanceof Short)) {
            return (short)o;
        }
        if ((o instanceof Byte)) {
            return (byte)o;
        }
        return (int)o;
    }
    
    static Integer boxInteger(int i) {
        return Integer.valueOf(i);
    }
    
    static Byte boxByte(byte b) {
        return Byte.valueOf(b);
    }
    
    static Short boxShort(short s) {
        return Short.valueOf(s);
    }
    
    static Boolean boxBoolean(boolean b) {
        return Boolean.valueOf(b);
    }
    
    static Character boxCharacter(char c) {
        return Character.valueOf(c);
    }
    
    static Long boxLong(long l) {
        return Long.valueOf(l);
    }
    
    static Float boxFloat(float f) {
        return Float.valueOf(f);
    }
    
    static Double boxDouble(double d) {
        return Double.valueOf(d);
    }
    
    private static MethodType boxType(Wrapper wrapper) {
        return MethodType.methodType(wrapper.wrapperType(), wrapper.primitiveType());
    }
    
    public static MethodHandle boxExact(Wrapper obj) {
        ValueConversions.WrapperCache wrapperCache = ValueConversions.BOX_CONVERSIONS[0];
        MethodHandle value = wrapperCache.get(obj);
        if (value != null) {
            return value;
        }
        String string = "box" + obj.wrapperSimpleName();
        MethodType boxType = boxType(obj);
        MethodHandle static1;
        try {
            static1 = ValueConversions.IMPL_LOOKUP.findStatic(ValueConversions.THIS_CLASS, string, boxType);
        }
        catch (ReflectiveOperationException ex) {
            static1 = null;
        }
        if (static1 != null) {
            return wrapperCache.put(obj, static1);
        }
        throw new IllegalArgumentException("cannot find box adapter for " + obj);
    }
    
    static void ignore(Object o) {
    }
    
    static void empty() {
    }
    
    static Object zeroObject() {
        return null;
    }
    
    static int zeroInteger() {
        return 0;
    }
    
    static long zeroLong() {
        return 0L;
    }
    
    static float zeroFloat() {
        return 0.0f;
    }
    
    static double zeroDouble() {
        return 0.0;
    }
    
    public static MethodHandle zeroConstantFunction(Wrapper obj) {
        ValueConversions.WrapperCache wrapperCache = ValueConversions.CONSTANT_FUNCTIONS[0];
        MethodHandle methodHandle = wrapperCache.get(obj);
        if (methodHandle != null) {
            return methodHandle;
        }
        MethodType methodType = MethodType.methodType(obj.primitiveType());
        switch (obj) {
            case VOID: {
                methodHandle = ValueConversions.EMPTY;
                break;
            }
            case OBJECT:
            case INT:
            case LONG:
            case FLOAT:
            case DOUBLE: {
                try {
                    methodHandle = ValueConversions.IMPL_LOOKUP.findStatic(ValueConversions.THIS_CLASS, "zero" + obj.wrapperSimpleName(), methodType);
                }
                catch (ReflectiveOperationException ex) {
                    methodHandle = null;
                }
                break;
            }
        }
        if (methodHandle != null) {
            return wrapperCache.put(obj, methodHandle);
        }
        if (obj.isSubwordOrInt() && obj != Wrapper.INT) {
            return wrapperCache.put(obj, MethodHandles.explicitCastArguments(zeroConstantFunction(Wrapper.INT), methodType));
        }
        throw new IllegalArgumentException("cannot find zero constant for " + obj);
    }
    
    public static MethodHandle ignore() {
        return ValueConversions.IGNORE;
    }
    
    public static MethodHandle cast() {
        return ValueConversions.CAST_REFERENCE;
    }
    
    static float doubleToFloat(double n) {
        return (float)n;
    }
    
    static long doubleToLong(double n) {
        return (long)n;
    }
    
    static int doubleToInt(double n) {
        return (int)n;
    }
    
    static short doubleToShort(double n) {
        return (short)(int)n;
    }
    
    static char doubleToChar(double n) {
        return (char)(int)n;
    }
    
    static byte doubleToByte(double n) {
        return (byte)(int)n;
    }
    
    static boolean doubleToBoolean(double n) {
        return toBoolean((byte)(int)n);
    }
    
    static double floatToDouble(float n) {
        return (double)n;
    }
    
    static long floatToLong(float n) {
        return (long)n;
    }
    
    static int floatToInt(float n) {
        return (int)n;
    }
    
    static short floatToShort(float n) {
        return (short)(int)n;
    }
    
    static char floatToChar(float n) {
        return (char)(int)n;
    }
    
    static byte floatToByte(float n) {
        return (byte)(int)n;
    }
    
    static boolean floatToBoolean(float n) {
        return toBoolean((byte)(int)n);
    }
    
    static double longToDouble(long n) {
        return (double)n;
    }
    
    static float longToFloat(long n) {
        return (float)n;
    }
    
    static int longToInt(long n) {
        return (int)n;
    }
    
    static short longToShort(long n) {
        return (short)(int)n;
    }
    
    static char longToChar(long n) {
        return (char)(int)n;
    }
    
    static byte longToByte(long n) {
        return (byte)(int)n;
    }
    
    static boolean longToBoolean(long n) {
        return toBoolean((byte)(int)n);
    }
    
    static double intToDouble(int n) {
        return (double)n;
    }
    
    static float intToFloat(int n) {
        return (float)n;
    }
    
    static long intToLong(int n) {
        return (long)n;
    }
    
    static short intToShort(int n) {
        return (short)n;
    }
    
    static char intToChar(int n) {
        return (char)n;
    }
    
    static byte intToByte(int n) {
        return (byte)n;
    }
    
    static boolean intToBoolean(int n) {
        return toBoolean((byte)n);
    }
    
    static double shortToDouble(short n) {
        return (double)n;
    }
    
    static float shortToFloat(short n) {
        return (float)n;
    }
    
    static long shortToLong(short n) {
        return (long)n;
    }
    
    static int shortToInt(short n) {
        return n;
    }
    
    static char shortToChar(short n) {
        return (char)n;
    }
    
    static byte shortToByte(short n) {
        return (byte)n;
    }
    
    static boolean shortToBoolean(short n) {
        return toBoolean((byte)n);
    }
    
    static double charToDouble(char c) {
        return (double)c;
    }
    
    static float charToFloat(char c) {
        return (float)c;
    }
    
    static long charToLong(char c) {
        return (long)c;
    }
    
    static int charToInt(char c) {
        return c;
    }
    
    static short charToShort(char c) {
        return (short)c;
    }
    
    static byte charToByte(char c) {
        return (byte)c;
    }
    
    static boolean charToBoolean(char c) {
        return toBoolean((byte)c);
    }
    
    static double byteToDouble(byte b) {
        return (double)b;
    }
    
    static float byteToFloat(byte b) {
        return (float)b;
    }
    
    static long byteToLong(byte b) {
        return (long)b;
    }
    
    static int byteToInt(byte b) {
        return b;
    }
    
    static short byteToShort(byte b) {
        return (short)b;
    }
    
    static char byteToChar(byte b) {
        return (char)b;
    }
    
    static boolean byteToBoolean(byte b) {
        return toBoolean(b);
    }
    
    static double booleanToDouble(boolean b) {
        return (double)fromBoolean(b);
    }
    
    static float booleanToFloat(boolean b) {
        return (float)fromBoolean(b);
    }
    
    static long booleanToLong(boolean b) {
        return (long)fromBoolean(b);
    }
    
    static int booleanToInt(boolean b) {
        return fromBoolean(b);
    }
    
    static short booleanToShort(boolean b) {
        return (short)fromBoolean(b);
    }
    
    static char booleanToChar(boolean b) {
        return (char)fromBoolean(b);
    }
    
    static byte booleanToByte(boolean b) {
        return fromBoolean(b);
    }
    
    static boolean toBoolean(byte b) {
        return (b & 0x1) != 0x0;
    }
    
    static byte fromBoolean(boolean b) {
        return b;
    }
    
    public static MethodHandle convertPrimitive(Wrapper wrapper, Wrapper wrapper2) {
        ValueConversions.WrapperCache wrapperCache = ValueConversions.CONVERT_PRIMITIVE_FUNCTIONS[wrapper.ordinal()];
        MethodHandle value = wrapperCache.get(wrapper2);
        if (value != null) {
            return value;
        }
        Class<Object> primitiveType = wrapper.primitiveType();
        Class<Object> primitiveType2 = wrapper2.primitiveType();
        MethodType methodType = MethodType.methodType(primitiveType2, primitiveType);
        MethodHandle detailMessage;
        if (wrapper == wrapper2) {
            detailMessage = MethodHandles.identity(primitiveType);
        }
        else {
            assert primitiveType.isPrimitive() && primitiveType2.isPrimitive();
            try {
                detailMessage = ValueConversions.IMPL_LOOKUP.findStatic(ValueConversions.THIS_CLASS, primitiveType.getSimpleName() + "To" + capitalize(primitiveType2.getSimpleName()), methodType);
            }
            catch (ReflectiveOperationException ex) {
                detailMessage = null;
            }
        }
        if (detailMessage != null) {
            assert detailMessage.type() == methodType : detailMessage;
            return wrapperCache.put(wrapper2, detailMessage);
        }
        else {
            throw new IllegalArgumentException("cannot find primitive conversion function for " + primitiveType.getSimpleName() + " -> " + primitiveType2.getSimpleName());
        }
    }
    
    public static MethodHandle convertPrimitive(Class<?> clazz, Class<?> clazz2) {
        return convertPrimitive(Wrapper.forPrimitiveType(clazz), Wrapper.forPrimitiveType(clazz2));
    }
    
    private static String capitalize(String s) {
        return Character.toUpperCase(s.charAt(0)) + s.substring(1);
    }
    
    private static InternalError newInternalError(String message, Throwable cause) {
        return new InternalError(message, cause);
    }
    
    private static InternalError newInternalError(Throwable cause) {
        return new InternalError(cause);
    }
    
    static {
        THIS_CLASS = ValueConversions.class;
        IMPL_LOOKUP = MethodHandles.lookup();
        UNBOX_CONVERSIONS = newWrapperCaches(4);
        ZERO_INT = Integer.valueOf(0);
        ONE_INT = Integer.valueOf(1);
        BOX_CONVERSIONS = newWrapperCaches(1);
        CONSTANT_FUNCTIONS = newWrapperCaches(2);
        try {
            MethodType genericMethodType = MethodType.genericMethodType(1);
            MethodType changeReturnType = genericMethodType.changeReturnType(Void.TYPE);
            CAST_REFERENCE = ValueConversions.IMPL_LOOKUP.findVirtual(Class.class, "cast", genericMethodType);
            IGNORE = ValueConversions.IMPL_LOOKUP.findStatic(ValueConversions.THIS_CLASS, "ignore", changeReturnType);
            EMPTY = ValueConversions.IMPL_LOOKUP.findStatic(ValueConversions.THIS_CLASS, "empty", changeReturnType.dropParameterTypes(0, 1));
        }
        catch (NoSuchMethodException | IllegalAccessException ex) {
            throw newInternalError("uncaught exception", ex);
        }
        CONVERT_PRIMITIVE_FUNCTIONS = newWrapperCaches(Wrapper.values().length);
    }
    
    private static class WrapperCache
    {
        private final EnumMap<Wrapper, MethodHandle> map;
        
        private WrapperCache() {
            this.map = new EnumMap(Wrapper.class);
        }
        
        public MethodHandle get(Wrapper key) {
            return (MethodHandle)this.map.get(key);
        }
        
        public synchronized MethodHandle put(Wrapper key, MethodHandle value) {
            MethodHandle methodHandle = (MethodHandle)this.map.putIfAbsent(key, value);
            if (methodHandle != null) {
                return methodHandle;
            }
            return value;
        }
    }
}
